@@ -18,6 +18,9 @@
 // Copyright (C) 2010 Carlos Garcia Campos <carlosgc@gnome.org>
 // Copyright (C) 2012, 2013 Adrian Johnson <ajohnson@redneon.com>
 // Copyright (C) 2013 Thomas Freitag <Thomas.Freitag@alfa.de>
+// Copyright (C) 2023 g10 Code GmbH, Author: Sune Stolborg Vuorela <sune@vuorela.dk>
+// Copyright (C) 2023 Even Rouault <even.rouault@mines-paris.org>
+// Copyright (C) 2023 Oliver Sander <oliver.sander@tu-dresden.de>
 //
 // To see a description of the changes please see the Changelog file that
 // came with your tarball or type make ChangeLog if you are building from git
@@ -33,6 +36,7 @@
 #include <cctype>
 #include "Lexer.h"
 #include "Error.h"
+#include "UTF.h"
 #include "XRef.h"
 
 //------------------------------------------------------------------------
@@ -163,7 +167,7 @@ Object Lexer::getObj(int objNum)
     int xi;
     long long xll = 0;
     double xf = 0, scale;
-    GooString *s;
+    std::unique_ptr<GooString> s;
     int n, m;
 
     // skip whitespace and comments
@@ -389,7 +393,7 @@ Object Lexer::getObj(int objNum)
             if (c2 != EOF) {
                 if (n == tokBufSize) {
                     if (!s) {
-                        s = new GooString(tokBuf, tokBufSize);
+                        s = std::make_unique<GooString>(tokBuf, tokBufSize);
                     } else {
                         s->append(tokBuf, tokBufSize);
                     }
@@ -402,7 +406,7 @@ Object Lexer::getObj(int objNum)
                         if (newObjNum != objNum) {
                             error(errSyntaxError, getPos(), "Unterminated string");
                             done = true;
-                            delete s;
+                            s.reset();
                             n = -2;
                         }
                     }
@@ -413,11 +417,15 @@ Object Lexer::getObj(int objNum)
         } while (!done);
         if (n >= 0) {
             if (!s) {
-                s = new GooString(tokBuf, n);
+                s = std::make_unique<GooString>(tokBuf, n);
             } else {
                 s->append(tokBuf, n);
             }
-            return Object(s);
+            // Check utf8
+            if (isUtf8WithBom(s->toStr())) {
+                s = std::make_unique<GooString>(utf8ToUtf16WithBom(s->toStr()));
+            }
+            return Object(s.release());
         } else {
             return Object(objEOF);
         }
@@ -464,8 +472,13 @@ Object Lexer::getObj(int objNum)
             } else if (n == tokBufSize) {
                 error(errSyntaxError, getPos(), "Warning: name token is longer than what the specification says it can be");
                 *p = c;
-                s = new GooString(tokBuf, n);
+                s = std::make_unique<GooString>(tokBuf, n);
             } else {
+                // Somewhat arbitrary threshold
+                if (unlikely(n == 1024 * 1024)) {
+                    error(errSyntaxError, getPos(), "Error: name token is larger than 1 MB. Suspicion of hostile file. Stopping parsing");
+                    return Object(objEOF);
+                }
                 s->append((char)c);
             }
         }
@@ -474,7 +487,6 @@ Object Lexer::getObj(int objNum)
             return Object(objName, tokBuf);
         } else {
             Object obj(objName, s->c_str());
-            delete s;
             return obj;
         }
         break;
@@ -525,7 +537,7 @@ Object Lexer::getObj(int objNum)
                     if (++m == 2) {
                         if (n == tokBufSize) {
                             if (!s) {
-                                s = new GooString(tokBuf, tokBufSize);
+                                s = std::make_unique<GooString>(tokBuf, tokBufSize);
                             } else {
                                 s->append(tokBuf, tokBufSize);
                             }
@@ -540,14 +552,17 @@ Object Lexer::getObj(int objNum)
                 }
             }
             if (!s) {
-                s = new GooString(tokBuf, n);
+                s = std::make_unique<GooString>(tokBuf, n);
             } else {
                 s->append(tokBuf, n);
             }
             if (m == 1) {
                 s->append((char)(c2 << 4));
             }
-            return Object(s);
+            if (isUtf8WithBom(s->toStr())) {
+                s = std::make_unique<GooString>(utf8ToUtf16WithBom(s->toStr()));
+            }
+            return Object(s.release());
         }
         break;
 
